"""Dataset registry for ktch.datasets module."""

# This file is auto-generated by:
#   uv run python scripts/update_registry.py
# Do not edit manually. Edit registry.toml instead.

# Base URL for remote datasets
BASE_URL = "https://pub-c1d6dba6c94843f88f0fd096d19c0831.r2.dev"

# Per-dataset registry: {dataset_name: {version: {filename: sha256_hash}}}
dataset_registry = {
    "image_passiflora_leaves": {
        "1": {
            "image_passiflora_leaves.zip": "d23b22252d0baf3b60ef1986e6518f0550e2e1aa81f677b5991938ab65e2e45b",
        },
        "2": {
            "image_passiflora_leaves.zip": "1280c46cda7b273a006fb20e90b5ef4c5098846ccacc6fba805244423c0392ca",
        },
    },
    "surface_leaf_bending": {
        "1": {
            "surface_leaf_bending.zip": "4d6792375e08839eaa7eb8107401ea60ed4d4b47f4677ac5d6b5d590670fc1e0",
        },
    },
}

# Default dataset versions for the current ktch release.
# Updated at ktch release time to pin the recommended dataset version.
default_versions = {
    "image_passiflora_leaves": "2",
    "surface_leaf_bending": "1",
}
