"""
The :mod:`ktch.landmark` module implements landmark-based morphometrics.
"""

# Copyright 2020 Koji Noshita
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ._procrustes_analysis import (
    GeneralizedProcrustesAnalysis,
    centroid_size,
    combine_landmarks_and_curves,
    define_curve_sliders,
)

__all__ = [
    "GeneralizedProcrustesAnalysis",
    "centroid_size",
    "combine_landmarks_and_curves",
    "define_curve_sliders",
]
